/*
* This file is part of the XKCD Velociraptor Simulation originally written by Matthew Beckler, Grant Miller, and John Saxton.
* 
* the XKCD Velociraptor Simulation is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
* 
* The XKCD Velociraptor Simulation is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with the XKCD Velociraptor Simulation; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
**/

import javax.swing.*;
import java.awt.*;
import java.io.*;
import java.awt.event.*;
import javax.imageio.*;
import java.awt.image.*;
import java.text.*;

public class VisualizeRaptor extends JFrame implements MouseListener
{
	private DataPoint[] dataPoints;
	private int currentDataPoint = 0;
	private boolean trails = false;
	//if trails is true, all the previous steps are also drawn, making a pretty curve
	
	public VisualizeRaptor(DataPoint[] dataPoints, boolean trails)
	{
		super("Raptor Visualization");
		this.dataPoints = dataPoints;
		currentDataPoint = 300;
		this.trails = trails;
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		setSize(new Dimension(600, 630));
		addMouseListener(this);
		setVisible(true);
		repaint();
	}	

	public void paint(Graphics g)
	{
		paintThisOne(g, dataPoints, currentDataPoint, trails);

		//if you like, you can un-comment the following block, and it will try to print out the character's velocities, as determined from the previous step's position, by dividing the change in position by the timestep. We didn't fully test this out, but it appears to be accurate.

		
		if (currentDataPoint > 0) //can't do the calculation with only one point, so we must wait until we get to at least point 1
		{
			//we can calculate velocity
			DataPoint dp1 = dataPoints[currentDataPoint - 1];
			DataPoint dp2 = dataPoints[currentDataPoint];
			double dt = dp2.time - dp1.time;
			System.out.println("Time = " + dp2.time);
			System.out.println("  Victim Velocity: " + Math.sqrt(Math.pow(dp2.runnerX - dp1.runnerX, 2) + Math.pow(dp2.runnerY - dp1.runnerY, 2)) / dt);
			System.out.println("  Injured Raptor Velocity: " + Math.sqrt(Math.pow(dp2.injuredRaptorX - dp1.injuredRaptorX, 2) + Math.pow(dp2.injuredRaptorY - dp1.injuredRaptorY, 2)) / dt);
			System.out.println("  Left Raptor Velocity: " + Math.sqrt(Math.pow(dp2.leftRaptorX - dp1.leftRaptorX, 2) + Math.pow(dp2.leftRaptorY - dp1.leftRaptorY, 2)) / dt);
			System.out.println("  Right Raptor Velocity: " + Math.sqrt(Math.pow(dp2.rightRaptorX - dp1.rightRaptorX, 2) + Math.pow(dp2.rightRaptorY - dp1.rightRaptorY, 2)) / dt);
		}
		
	}
	
	//in the interactive GUI version, this method detects if the user has clicked in a button, and responds appropriately
	public void mouseClicked(MouseEvent event)
	{
		int x = event.getX();
		int y = event.getY();
		//System.out.println("Clicked: (" + x + ", " + y + ")");
		if (x >= 400 && x <= 430 && y >= 605 && y <= 620 && currentDataPoint > 0)
		{
			currentDataPoint -= 1;
			repaint();
		}

		if (x >= 440 && x <= 470 && y >= 605 && y<= 620 && currentDataPoint < dataPoints.length - 1)
		{
			currentDataPoint += 1;
			repaint();
		}
		
		if (x >= 480 && x <= 560 && y >= 605 && y<= 620)
		{
			trails = !trails;
			repaint();
		}

	}
	public void mouseReleased(MouseEvent e){}
	public void mousePressed(MouseEvent e){}
	public void mouseEntered(MouseEvent e){}
	public void mouseExited(MouseEvent e){}

	//this method translates from the problem's standard coordinate system, into the computer's pixel coordinate system
	//Basically, it says that the screen area (600px by 600px) can represent a 35m by 35m area, so (-17.5, 17.5) in each direction
	//c2p stands for coordinate2pixel
	public static int c2p(double coordinate)
	{
		return (int)((coordinate / 35.0) * 600.0) + 300;
	}

	//this has been extracted out of the paint method, to allow for drawing to files instead of to the screen
	public static void paintThisOne(Graphics g, DataPoint[] dataPoints, int currentDataPoint, boolean trails)
	{
		//feel free to change the colors of the victim and raptors. Personally, I though raptors should be green, and I happened to be wearing a blue shirt the day I wrote this class, so there :-p
		Color white = new Color(255, 255, 255);
		Color gray = new Color(220, 220, 220);
		Color black = new Color(0, 0, 0);
		Color victim = new Color(50, 50, 200);
		Color raptor = new Color(0, 200, 0);

		DecimalFormat df1 = new DecimalFormat("0.00");
		//used to format the current time, because seriously, with a time step of 0.01 second, who really needs 9 digits of (in-)accuracy?


		//clear the screen
		g.setColor(white);
		g.fillRect(0, 0, 600, 630);
		
		//draw axes
		g.setColor(gray);
		g.drawLine(0, 300, 600, 300);
		g.drawLine(300, 0, 300, 600);

		//draw tic-marks on the axes every 5 meters
		int ticSize = 10;
		for (int x = -15; x < 20; x += 5)
		{
			g.drawLine(c2p(x), 300 + ticSize, c2p(x), 300 - ticSize);
			if (x != 0) // it's hard to draw the 0 mark with the same rule as the normal tic's
			{
				g.drawString(x + "m", (c2p(x) - 4) - ((x + "m").length() * 2), 300 + ticSize + 15);
			}
		}
		
		for (int y = -15; y < 20; y += 5)
		{
			g.drawLine(300 + ticSize, c2p(y), 300 - ticSize, c2p(y));
			if (y != 0) // it's hard to draw the 0 mark with the same rule as the normal tic's
			{
				g.drawString(y + "m", 300 + ticSize + 6, (c2p(-y) + 4));
			}
		}
		
		//create the information area
		g.setColor(black);
		g.drawLine(0, 600, 600, 600);
		g.drawString("Viewing Step " + currentDataPoint + " of " + dataPoints.length + " at time = " + df1.format(dataPoints[currentDataPoint].time), 10, 620);
		
		//draw buttons
		g.drawRect(400, 605, 30, 15);
		g.drawString("Prev", 402, 618);
		
		g.drawRect(440, 605, 30, 15);
		g.drawString("Next", 442, 618);
		
		g.drawRect(480, 605, 80, 15);
		g.drawString("Toggle Trails", 482, 618);
		
		for (int i = 0; i < currentDataPoint; i++)
		{
			if (!trails)
			{
				//if !trails, then just skip to the current point, and the loop will only execute once :-)
				i = currentDataPoint;
			}

			DataPoint dp = dataPoints[i];

			//draw the victim
			g.setColor(victim);
			g.drawRect(c2p(dp.runnerX), c2p(-dp.runnerY), 1, 1);

			//draw the raptors			
			g.setColor(raptor);
			g.drawRect(c2p(dp.injuredRaptorX), c2p(-dp.injuredRaptorY), 1, 1);
			g.drawRect(c2p(dp.leftRaptorX), c2p(-dp.leftRaptorY), 1, 1);
			g.drawRect(c2p(dp.rightRaptorX), c2p(-dp.rightRaptorY), 1, 1);
		}
	}
	
	public static void main(String[] arrrgs)
	{
		boolean toFiles = false; //default to screen
		boolean trails = false; //default to noTrails
		DataPoint[] dataPoints = null;
		//first, check for proper arrrguments
		//Arrrguments are optional, it defaults to displaying toScreen with noTrails

		System.out.println("USAGE: java VisualizeRaptor [toScreen,toFiles] [yesTrails,noTrails]");
		System.out.println("[toScreen,toFiles] is an optional argument, that determines where the images are drawn. - Defaults to 'toScreen'");
		System.out.println("[yesTrails,noTrails] is an optional argument, that determines if trails are drawn. - Defaults to 'noTrails'");
		System.out.println();


		if (arrrgs.length >= 1 && arrrgs[0].equalsIgnoreCase("toFiles"))
		{
			toFiles = true;
		}

		if (arrrgs.length >= 1 && arrrgs[0].equalsIgnoreCase("yesTrails"))
		{
			trails = true;
		}

		if (arrrgs.length == 2 && arrrgs[1].equalsIgnoreCase("yesTrails"))
		{
			trails = true;
		}

		//for either screen or files, we still need to read in the data file, stored in output.csv
		//data is stored in the csv file in the following format:
		//time, runnerX, runnerY, injuredRaptorX, injuredRaptorY, leftRaptorX, leftRaptorY, rightRaptorX, rightRaptorY
		try
		{
			BufferedReader reader = new BufferedReader(new FileReader("output.csv"));
			int numberOfLines = 0;
			while (reader.readLine() != null)
			{
				numberOfLines++;
			}
			
			System.out.println("Found " + numberOfLines + " lines in input file");
			
			reader.close();
			//yes, yes, yes, before you ask, this really is a hack. I didn't really want to mess with a more advanced data type, and have to worry about type-casting things. I've heard that the 'generics' thing in java 1.5 is supposed to fix that inconvenience...
			reader = new BufferedReader(new FileReader("output.csv"));
			String[] values;
			dataPoints = new DataPoint[numberOfLines];
			
			for (int lineNumber = 0; lineNumber < numberOfLines; lineNumber++)
			{
				values = reader.readLine().split(", ");
				dataPoints[lineNumber] = new DataPoint(Double.parseDouble(values[0]),
										Double.parseDouble(values[1]),
										Double.parseDouble(values[2]),
										Double.parseDouble(values[3]),
										Double.parseDouble(values[4]),
										Double.parseDouble(values[5]),
										Double.parseDouble(values[6]),
										Double.parseDouble(values[7]),
										Double.parseDouble(values[8]));
			}
		}
		catch (FileNotFoundException e)
		{
			System.err.println("Please make sure you have already run RaptorEvaluate, and that the file 'output.csv' exists!");
			System.exit(1);
		}
		catch (IOException e)
		{
			System.err.println("An IO exception has occured");
			e.printStackTrace();
			System.exit(1);
		}

		//at this point, dataPoints[] has been properly initialized, and we are ready to either:
		// 1. Display the interactive GUI
		// 2. Write each frame into the 'images/' folder

		if (arrrgs.length >= 1 && arrrgs[0].equalsIgnoreCase("toFiles"))
		{
			int currentDataPoint = 0;

			for (int i = 0; i < dataPoints.length; i++)
			{
				System.out.println("Painting point: " + i);
				currentDataPoint = i;
			
				try
				{
					BufferedImage image = new BufferedImage(600, 630, BufferedImage.TYPE_INT_RGB);
					Graphics gOut = image.createGraphics();
					paintThisOne(gOut, dataPoints, currentDataPoint, trails);
					//okay, a bit of explanation about the filename...adding 10000 to the data point pretty much fixes the ordering problems that you sometimes get with files named, for example, raptor1.png, raptor2.png,..., raptor10.png, raptor100.png, etc
					//this way, the files are named raptor10001.png, raptor10002.png, etc
					ImageIO.write(image, "png", new File("images/raptor" + (currentDataPoint + 10000) + ".png"));
				}
				catch (Exception e)
				{
					e.printStackTrace();
				}
			}
		}
		else
		{ // must have entered 'toScreen'
			//create an instance of the class VisualizeRaptor
			VisualizeRaptor vr = new VisualizeRaptor(dataPoints, trails);
		}
	}
}

class DataPoint
{
	public double time;
	public double runnerX;
	public double runnerY;
	public double injuredRaptorX;
	public double injuredRaptorY;
	public double leftRaptorX;
	public double leftRaptorY;
	public double rightRaptorX;
	public double rightRaptorY;
	
	public DataPoint(double time, double runnerX, double runnerY, double injuredRaptorX, double injuredRaptorY, double leftRaptorX, double leftRaptorY, double rightRaptorX, double rightRaptorY)
	{
		this.time = time;
		this.runnerX = runnerX;
		this.runnerY = runnerY;
		this.injuredRaptorX = injuredRaptorX;
		this.injuredRaptorY = injuredRaptorY;
		this.leftRaptorX = leftRaptorX;
		this.leftRaptorY = leftRaptorY;
		this.rightRaptorX = rightRaptorX;
		this.rightRaptorY = rightRaptorY;
	}
}
